package com.bergerkiller.bukkit.nolagg.meta;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import net.minecraft.server.EntityItem;
import net.minecraft.server.WorldServer;

import org.bukkit.Chunk;
import org.bukkit.Location;
import org.bukkit.World;
import org.bukkit.craftbukkit.CraftChunk;
import org.bukkit.craftbukkit.CraftWorld;
import org.bukkit.entity.Item;
import org.bukkit.inventory.ItemStack;

import com.bergerkiller.bukkit.common.utils.CommonUtil;
import com.bergerkiller.bukkit.common.utils.EntityUtil;
import com.bergerkiller.bukkit.common.utils.MathUtil;
import com.bergerkiller.bukkit.nolagg.ItemHandler;
import com.bergerkiller.bukkit.nolagg.NoLagg;

public class ChunkMetaData {
	public static int unloadDelay;
	public static boolean useUnloadDelay;
	
	public static ChunkMetaData get(Chunk chunk) {
		return get(chunk.getWorld(), chunk.getX(), chunk.getZ());
	}
	public static ChunkMetaData get(final World world, final int x, final int z) {
		return WorldMetaData.get(world).getChunk(x, z);
	}
	public static ChunkMetaData get(final Location location) {
		return WorldMetaData.get(location.getWorld()).getChunk(location);
	}
	
	protected ChunkMetaData(final WorldMetaData world, final int x, int z) {
		this.world = world;
		this.spawnedItems = new ArrayList<Item>();
		this.hiddenItems = new ArrayList<Item>();
		this.x = x;
		this.z = z;
		this.add();
	}
	private final WorldMetaData world;
	private final int x, z;
	private final ArrayList<Item> spawnedItems;
	private final ArrayList<Item> hiddenItems;
	private boolean respawnItemsNeeded = false;
	
	public long lastAccessTime;
	private short unloadCounter = 0;
	
	public WorldMetaData getMetaWorld() {
		return this.world;
	}
	public CraftWorld getWorld() {
		return this.world.world;
	}
	public WorldServer getNativeWorld() {
		return this.world.getNative();
	}
	public net.minecraft.server.Chunk getNativeChunk() {
		return this.getChunk().getHandle();
	}
	public CraftChunk getChunk() {
		return (CraftChunk) this.getWorld().getChunkAt(this.x, this.z);
	}
	
	public int getX() {
		return this.x;
	}
	public int getZ() {
		return this.z;
	}
	
	public boolean isLoaded() {
		return this.getWorld().isChunkLoaded(this.getX(), this.getZ());
	}
	
	public void touch() {
		this.lastAccessTime = System.currentTimeMillis();
	}
	
	public void markNotUnloadable() {
		this.unloadCounter = 5; //ignore this chunk for 5 update cycles
	}
	
	public boolean isExpired() {
		return this.isExpired(System.currentTimeMillis() - unloadDelay);
	}
	public boolean isExpired(final long time) {
		if (!useUnloadDelay) return true;
		return this.lastAccessTime <= time;
	}
	
	public boolean isPlayerNear() {
		final int bx = this.getX() << 4;
		final int bz = this.getZ() << 4;
		net.minecraft.server.Entity e;
		for (Object o : this.getNativeWorld().players) {
			e = (net.minecraft.server.Entity) o;
			if (EntityUtil.isNearBlock(e, bx, bz, CommonUtil.blockView)) return true;
		}
		return false;
	}
	
	public boolean isNearSpawnChunk(WorldServer world) {
		return this.isNearChunk(WorldMetaData.getSpawnChunkX(world), WorldMetaData.getSpawnChunkZ(world));
	}
	public boolean isNearChunk(int cx, int cz) {
		return Math.abs(cx - this.x) <= 12 && Math.abs(cz - this.z) <= 12;
	}
	
	public boolean isUnloadable() {
		return this.isUnloadable(true);
	}
	public boolean isUnloadable(boolean checkCounter) {
		WorldServer world = this.getNativeWorld();
		if (world.savingDisabled) return false;
		if (checkCounter && this.unloadCounter > 0) {
			--unloadCounter;
			return false;
		} else {
			return !world.keepSpawnInMemory || !this.isNearSpawnChunk(world);
		}
	}

	public String toString() {
		return "[" + this.getX() + "/" + this.getZ() + "/" + this.getWorld().getName() + "]";
	}
	public long toLongCoordinate() {
		return MathUtil.toLong(this.x, this.z);
	}
			
	public List<Item> getSpawnedItems() {
		synchronized (this.spawnedItems) {
			return this.spawnedItems;
		}
	}
	public int getHiddenItemCount() {
		synchronized (this.spawnedItems) {
			return this.hiddenItems.size();
		}
	}
	public void clearItems() {
		synchronized (this.spawnedItems) {
			this.hiddenItems.clear();
			this.spawnedItems.clear();
		}
	}
	public boolean removeSpawnedItem(Item item) {
		synchronized (this.spawnedItems) {
			if (this.spawnedItems.remove(item)) {
				this.respawnItemsNeeded = true;
				return true;
			} else {
				return false;
			}
		}
	}
	private void spawnInChunk() {
		synchronized (this.spawnedItems) {
			for (int i = this.spawnedItems.size(); i <= ItemHandler.maxItemsPerChunk && this.hiddenItems.size() > 0; i++) {
				ItemHandler.respawnItem(this.hiddenItems.remove(0));
			}
			this.respawnItemsNeeded = false;
		}
	}
	public boolean handleItemSpawn(Item item, boolean doIgnoreCheck) {
		synchronized (spawnedItems) {
			if (ItemHandler.ignoreSpawn || !NoLagg.bufferItems || ItemHandler.maxItemsPerChunk < 0) return true;
			if (ItemHandler.maxItemsPerChunk == 0) return false;
			if (doIgnoreCheck && ItemHandler.isIgnoredItem(item)) return true;
			int currentcount = this.spawnedItems.size();
			if (currentcount > ItemHandler.maxItemsPerChunk) {
				if (ItemHandler.formStacks) {
					//can we add to an existing stack or is a new item required?
					ItemStack main = item.getItemStack();
					int amount = main.getAmount();
					int max = main.getType().getMaxStackSize();
					boolean add = true;
					for (Item hiddenitem : this.hiddenItems) {
						ItemStack hiddenstack = hiddenitem.getItemStack();
						if (hiddenstack.getType() == main.getType() && hiddenstack.getDurability() == main.getDurability()) {
							int hidamount = hiddenstack.getAmount();
							if (hidamount < max) {
								if ((hidamount + amount) <= max) {
									hiddenstack.setAmount(hidamount + amount);
									add = false;
									break;
								} else {
									amount -= max - hidamount;
									hiddenstack.setAmount(max);
								}
							}
						}
					}
					if (add) {
						main.setAmount(amount);
						this.hiddenItems.add(item);
					}
				} else {
					this.hiddenItems.add(item);
				}
				return false;
			} else {
				this.spawnedItems.add(item);
				return true;
			}
		}
	}
		
	/*
	 * Called every so much ticks and during general events
	 */
	public void update() {
		//update spawned items
		if (NoLagg.bufferItems) {
			synchronized (this.spawnedItems) {
				boolean removed = false;
				Iterator<Item> spawned = this.spawnedItems.iterator();
				while (spawned.hasNext()) {
					if (spawned.next().isDead()) {
						spawned.remove();
						removed = true;
					}
				}
				if (removed || this.respawnItemsNeeded) this.spawnInChunk();
			}
		}
	}
	
	/*
	 * Called when this chunk meta data is first added
	 */
	@SuppressWarnings("rawtypes")
	public void add() {
		//load items
		net.minecraft.server.Chunk c = this.getNativeChunk();
		if (c.entitySlices != null) {
			Iterator it;
			for (int i = 0; i < c.entitySlices.length; i++) {
				it = ((List) c.entitySlices[i]).iterator();
				while (it.hasNext()) {
					net.minecraft.server.Entity e = (net.minecraft.server.Entity) it.next();
					if (e instanceof EntityItem && !this.handleItemSpawn((Item) e.getBukkitEntity(), true)) {
						//remove item
						e.dead = true;
					}
				}
			}
		}
	}
		
	/*
	 * Called when this chunk meta data is being removed (unloaded)
	 */
	public void remove() {
		this.remove(true);
	}
	protected void remove(final boolean removefromworldmeta) {
		//respawn items
		for (Item item : this.hiddenItems) ItemHandler.respawnItem(item);
		if (removefromworldmeta) this.world.removeChunk(this);
	}
	
	public boolean isChunk(Chunk chunk) {
		if (chunk.getWorld() != this.getWorld()) return false;
		if (chunk.getX() != this.getX()) return false;
		if (chunk.getZ() != this.getZ()) return false;
		return true;
	}

}
