package com.bergerkiller.bukkit.nolaggchunks;

import java.util.logging.Level;

import org.bukkit.Bukkit;
import org.bukkit.ChatColor;
import org.bukkit.command.CommandSender;
import org.bukkit.entity.Player;
import org.bukkit.permissions.PermissionDefault;
import org.bukkit.plugin.Plugin;

import com.bergerkiller.bukkit.common.MessageBuilder;
import com.bergerkiller.bukkit.common.PluginBase;
import com.bergerkiller.bukkit.common.config.FileConfiguration;
import com.bergerkiller.bukkit.common.utils.CommonUtil;
import com.bergerkiller.bukkit.common.utils.MathUtil;

public class NoLaggChunks extends PluginBase {

	public NoLaggChunks() {
		super(1838, 1938);
	}
	
	public static boolean isOreObfEnabled = false;
	public static boolean isOreToClientObfEnabled = false;
	public static boolean isSpoutEnabled = false;

	@Override
	public void updateDependency(Plugin plugin, String pluginName, boolean enabled) {
		if (pluginName.equals("Spout")) {
			isSpoutEnabled = enabled;
		} else if (pluginName.equals("Chunk Manager")) {
			if (enabled) {
				log(Level.WARNING, "Chunk Manager detected, NoLaggChunks has been disabled.");
				log(Level.WARNING, "Either disable Chunk Manager or disable NoLaggChunks.");
				Bukkit.getServer().getPluginManager().disablePlugin(this);
			}
		} else if (pluginName.equals("Orebfuscator")) {
			if (isOreObfEnabled = enabled) {
				log(Level.INFO, "Orebfuscation has been detected and will be used when sending chunks");
				log(Level.INFO, "Note that this may require you to set more threads used for sending!");
			}
		} else if (pluginName.equals("RawcriticsOreObfuscationPluginSpout")) {
			if (isOreToClientObfEnabled = enabled) {
				log(Level.INFO, "Ore To Client Obfuscation (Spout) has been detected and will be used when sending chunks");
				log(Level.INFO, "Note that this may require you to set more threads used for sending!");
			}
		}
	}

	public static NoLaggChunks plugin;
	
	
	public void enable() {
		plugin = this;
			
		this.register(NLCListener.class);
		this.register("sending");	
		
		FileConfiguration config = new FileConfiguration(this);
		config.load();
		
		config.setHeader("This is the configuration of NoLaggChunks\n");
		config.setHeader("minRate", "The minimum chunk sending rate (chunks/tick)");
		config.setHeader("maxRate", "The maximum chunk sending rate (chunks/tick)");
		config.setHeader("triggerRate", "The rate at which chunks are 'triggered' (refreshed, make visible) (trigger packets/tick)");
		config.setHeader("compressionThreadCount", "The amount of threads to use to compress the chunk packets (increase if it can't keep up)");
		ChunkSendQueue.minRate = config.get("minRate", 0.25);
		ChunkSendQueue.maxRate = config.get("maxRate", 2.00);
		ChunkSendQueue.globalTriggerRate = config.get("triggerRate", 0.5);
		ChunkCompressionThread.init(config.get("compressionThreadCount", 2));
		
		config.save();
		
		ChunkSendQueue.init();
	}
		
	public void disable() {
		ChunkSendQueue.deinit();
		ChunkCompressionThread.deinit();
	}

	@Override
	public void permissions() {
		this.loadPermission("nolaggchunks.info", PermissionDefault.OP, "The player can view sending information about himself");
	}

	@Override
	public boolean command(CommandSender sender, String command, String[] args) {
		if (command.equals("sending")) {
			double avgrate = MathUtil.round(ChunkSendQueue.getAverageRate(), 2);
			double compbus = MathUtil.round(ChunkSendQueue.compressBusyPercentage, 2);
			if (sender instanceof Player) {
				if (!sender.hasPermission("nolaggchunks.info")) {
					sender.sendMessage(ChatColor.RED + "You don't have permission to use this!");
					return true;
				}
				
				//show sending information of the player
				ChunkSendQueue queue = ChunkSendQueue.bind((Player) sender);
				if (queue != null) {
					MessageBuilder msg = new MessageBuilder();
					msg.green("You receive ").yellow(MathUtil.round(queue.getRate(), 2));
					msg.green(" chunks each tick (");
					msg.yellow(avgrate).yellow(" avg").green(")").newLine();
				    msg.green("You have received ");
				    int sent = CommonUtil.chunkArea - queue.getPendingSize();
				    double per = MathUtil.round((double) sent / (double) CommonUtil.chunkArea * 100.0, 2);
				    msg.yellow(sent).white("/").yellow(CommonUtil.chunkArea).green(" chunks (");
				    msg.yellow(per, "%").green(")").newLine();
				    msg.green("Your packet buffer is ").green(queue.getBufferLoadMsg()).green(" used");
				    msg.newLine().green("Chunk compression is ");
				    if (compbus > 70) {
				    	msg.red(compbus, "%");
				    } else if (compbus > 40) {
				    	msg.yellow(compbus, "%");
				    } else {
				    	msg.green(compbus, "%");
				    }
				    msg.green(" of the time busy");
				    
					msg.send(sender);
				} else {
					sender.sendMessage("An unknown error occured!");
				}
			} else {
				//show average sending info
				sender.sendMessage("Average chunk sending rate: " + avgrate + " chunks each tick");
				sender.sendMessage("Chunk compression is " + compbus + "% of the time busy");
			}
		}
		return true;
	}
	
}
